/**
 * AutoDAO - Generic DAO on steroids implementation for Java.
 * Copyright (C) 2008  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.autodao;

import java.io.Serializable;

/**
 * Dao factory. Implementations must discover all declared DAOs. Then, for all entities that do not have a declared DAO,
 * generic one is created (such as it can be represented as <code>net.sf.autodao.Dao&lt;Entity,PrimaryKey&gt;</code>.
 * DAO factories are useful when entity-agnostic code is written.
 * It is a error to have more then one DAO per entity.
 * Warning! API and implementation of dao factories is not stabilized yet. Use at your own risk.
 */
public interface DaoFactory {

  /**
   * Get dao for given entity class.
   * You won't be able to use this method type-safely if your entity is parameterized.
   * Example:
   * <pre>
   * public class Entity&lt;T&gt; extends PersistentEntity&lt;Long&gt; { ... };
   * EntityDao dao = factory.getDao(Entity.class)
   * or
   * Entity&lt;?&gt; e = ...
   * EntityDao dao = factory.getDao(e.getClass())
   * </pre>
   * The problem is caused by <a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6184881">Bug 6184881</a>
   * in Java Language Specification.
   * Workaround (will produce unchecked warning):
   * <pre>
   * EntityDao dao = factory.getDao((Class)Entity.class);
   * </pre>
   * @param klass entity class. cannot be <code>null</code>.
   * @param <E>   entity class.
   * @param <K>   PK class.
   * @param <D>   dao class.
   * @return dao. Never <code>null</code>.
   */
  <E extends PersistentEntity<K>, K extends Serializable, D extends Dao<E, K>> D getDao(
    final Class<? extends E> klass
  );

  /**
   * Get dao for given entity object.
   * @param entity entity. Cannot be <code>null</code>.
   * @param <E>    entity class.
   * @param <K>    PK class.
   * @param <D>    dao class.
   * @return dao. Never <code>null</code>.
   */
  <E extends PersistentEntity<K>, K extends Serializable, D extends Dao<E, K>> D getDao(final E entity);
}
