/**
 * AutoDAO - Generic DAO on steroids implementation for Java.
 * Copyright (C) 2008  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.autodao.hibernate;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import net.sf.autodao.Dao;
import net.sf.autodao.Limit;
import net.sf.autodao.Offset;
import net.sf.autodao.PersistentEntity;
import net.sf.autodao.SpecificationViolationException;
import net.sf.autodao.Utils;
import static net.sf.autodao.hibernate.HibernateUtils.getNamedQueryName;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.hibernate.SessionFactory;
import org.hibernate.engine.SessionFactoryImplementor;
import org.springframework.aop.framework.ProxyFactoryBean;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Required;
import static org.springframework.util.Assert.isAssignable;
import static org.springframework.util.Assert.isTrue;
import static org.springframework.util.Assert.notNull;

/**
 * Validating proxy factory.
 * NOT FOR PUBLIC USE.
 */
public final class HibernateDaoProxyFactoryBean extends ProxyFactoryBean implements InitializingBean {

  /** Log. Just log. */
  private static final Log LOG = LogFactory.getLog(HibernateDaoProxyFactoryBean.class);

  /** Hibernate session factory. Required to extract class metadata. */
  private SessionFactory sessionFactory;

  /** Entity type. */
  private Class<? extends PersistentEntity<?>> type;

  /** {@inheritDoc} */
  public void afterPropertiesSet() {
    //Do not trust anyone. Especialy yourself.
    notNull(this.type, "No entity type given");
    isAssignable(PersistentEntity.class, this.type, "Invalid entity type");
    isTrue(super.getProxiedInterfaces().length == 1, "Expected only single proxy interface");

    final Class<?> daoInterface = super.getProxiedInterfaces()[0];
    isAssignable(Dao.class, daoInterface, "Invalid DAO interface");
    this.validate(daoInterface);
  }

  public Class<? extends PersistentEntity<?>> getType() {
    return this.type;
  }

  /**
   * Validates given DAO interface.
   * @param daoInterface dao interface.
   */
  private void validate(final Class<?> daoInterface) {
    SessionFactoryImplementor sfi = null;
    if (this.sessionFactory == null) {
      LOG.warn("No session factory given. Some validations were disabled for DAO " + daoInterface.getName());
    } else {
      notNull(
        this.sessionFactory.getClassMetadata(this.type),
        this.type.getName() + " is not a persistent class (it is not registered in SessionFactory)"
      );
      //Have a better way? Tell me.
      try {
        sfi = (SessionFactoryImplementor) this.sessionFactory;
      } catch (final ClassCastException e) {
        LOG.warn(
          "Could not cast session factory " + this.sessionFactory + " to " + SessionFactoryImplementor.class.getName()
            + ". Some validations were disabled for DAO " + daoInterface.getName()
        );
      }
    }

    final Set<String> seen = new HashSet<String>();
    for (final Method m : daoInterface.getMethods()) {
      if (Utils.isFinder(m)) {
        isTrue(seen.add(m.getName()), "AutoDAO doesn't support method overloading. Offending method: " + m);
        if (sfi != null) {
          final String queryName = getNamedQueryName(this.sessionFactory, this.type, m);
          isTrue(sfi.getNamedQuery(queryName) != null || sfi.getNamedSQLQuery(queryName) != null,
            "No named query '" + queryName + "' found. Method " + m + " cannot work."
          );
        }
        validateLimitOffset(m);
      } else {
        if (Dao.class.isAssignableFrom(daoInterface)) {
          isTrue(Arrays.asList(Dao.class.getMethods()).contains(m),
            "Method " + m + " is neigher a finder method nor is implemented in AutoDAO"
          );
        } else {
          throw new IllegalArgumentException("Method " + m + " is not a finder method");
        }
      }
    }
  }

  private static void validateLimitOffset(final Method meth) {
    final boolean isSingleFind = Utils.isSingleFind(meth);
    final Integer limitIndex = Utils.getLimitParameterIndex(meth);
    final Integer offsetIndex = Utils.getOffsetParameterIndex(meth);
    for (int i = 0; i < meth.getParameterAnnotations().length; i++) {
      final Class<?> paramType = meth.getParameterTypes()[i];
      final boolean isOffset = isOffset(meth.getParameterAnnotations()[i]);
      final boolean isLimit = isLimit(meth.getParameterAnnotations()[i]);
      if (isSingleFind) {
        if (isLimit || isOffset) {
          throw new SpecificationViolationException(
            "Neither @Limit nor @Offset is allowed on single-object finders",
            meth
          );
        }
      } else {
        if ((isLimit || isOffset)
          && !(Integer.class.isAssignableFrom(paramType))
          && !(Integer.TYPE.isAssignableFrom(paramType))) {
          throw new SpecificationViolationException(
            "@Limit and @Offset argument type must be either int or java.lang.Integer",
            meth
          );
        }
        if (isLimit && limitIndex != i || isOffset && offsetIndex != i) {
          throw new SpecificationViolationException(
            "@Limit and @Offset can appear only on one parameter",
            meth
          );
        }
        if (!isLimit && !isOffset) {
          if (limitIndex != null && i > limitIndex || offsetIndex != null && i > offsetIndex) {
            throw new SpecificationViolationException(
              "There cannot be any parameters after @Limit and @Offset parameters",
              meth
            );
          }
        }
      }
    }
  }

  private static <T> boolean hasInstance(final T[] arr, final Class<? extends T> klass) {
    for (final T t : arr) {
      if (klass.isInstance(t)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Returns true if given annotations array contains {@link Limit}.
   * @param ann annotations.
   * @return {@link Limit} presence.
   */
  private static boolean isLimit(final Annotation[] ann) {
    return hasInstance(ann, Limit.class);
  }

  /**
   * Returns true if given annotations array contains {@link Offset}.
   * @param ann annotations.
   * @return {@link Offset} presence.
   */
  private static boolean isOffset(final Annotation[] ann) {
    return hasInstance(ann, Offset.class);
  }

  @Required
  public void setSessionFactory(final SessionFactory sessionFactory) {
    this.sessionFactory = sessionFactory;
  }

  @Required
  public void setType(final Class<? extends PersistentEntity<?>> type) {
    this.type = type;
  }
}
